﻿using MetroFramework.Forms;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using MetroFramework.Controls;
using MC_Launcher_v2.Business.Minecraft.Profiles;
using MC_Launcher_v2.Business.Data;
using System.Threading;
using System.IO;
using MC_Launcher_v2.Business;
using MC_Launcher_v2.Presentation;
using Newtonsoft.Json;
using MC_Launcher_v2.Business.Minecraft;
using System.Diagnostics;

namespace MC_Launcher_v2.Presentation
{
    public partial class pnlMain : MetroUserControl
    {
        #region VARIABLES
        private JavaProfile currentGameProfile;
        private List<FileHashEntry> remoteHashes;
        private List<FileHashEntry> localHashes;
        private List<string> folderWhiteList, folderBlackList;
        private static string gameDir = Path.Combine(Application.StartupPath, "game");
        #endregion

        #region PROPERTIES
        public EventHandler GameStarted { get; set; }
        public JavaProfile CurrentGameProfile
        {
            get { return currentGameProfile; }
            set
            {
                if (currentGameProfile != value)
                {
                    currentGameProfile = value;
                    cfgGameSave.Enabled = cfgGameDelete.Enabled = cfgGameBar.Enabled = cfgGameArgs.Enabled = cfgGameExecutable.Enabled = cfgGameName.Enabled = currentGameProfile != null;
                    UpdateVersionEntries();
                    Launcher.JavaProfiles.ForEach(x => { if (x != null) x.Active = x == currentGameProfile; });
                    if (currentGameProfile != null)
                    {
                        cfgGameBar.Value = currentGameProfile.Ram;
                        cfgGameArgs.Text = currentGameProfile.Args;
                        cfgGameExecutable.Text = currentGameProfile.Executable;
                        cfgGameName.Text = currentGameProfile.Name;
                        if (cfgGameVersion.Items.Contains(currentGameProfile.Version))
                            cfgGameVersion.SelectedItem = currentGameProfile.Version;
                        if (cfgGameLibVersion.Items.Contains(currentGameProfile.LibVersion))
                            cfgGameLibVersion.SelectedItem = currentGameProfile.LibVersion;
                    }
                    else
                    {
                        cfgGameBar.Value = 0;
                        cfgGameArgs.Text = "";
                        cfgGameExecutable.Text = "";
                        cfgGameName.Text = "";
                        cfgGameVersion.SelectedItem = null;
                        cfgGameLibVersion.SelectedItem = null;
                    }
                    UpdateProfileInfo();
                }
            }
        }
        #endregion

        #region CONSTRUCTORS
        public pnlMain()
        {
            InitializeComponent();
            remoteHashes = new List<FileHashEntry>();
            localHashes = new List<FileHashEntry>();
            folderWhiteList = SimpleJsonFile.FromFile<List<string>>("whitelist.json");
            folderBlackList = SimpleJsonFile.FromFile<List<string>>("blacklist.json");

            cfgGameBar.Maximum = JavaProfile.MAX_MEMORY;
            cfgGameBar.Value = cfgGameBar.Maximum / 2;
            cfgGameRamMax.Text = SizeToString(cfgGameBar.Maximum);
            cfgGameRamCurrent.Text = SizeToString(cfgGameBar.Value);

            cfgGameSave.Enabled = cfgGameDelete.Enabled = cfgGameBar.Enabled = cfgGameArgs.Enabled = cfgGameExecutable.Enabled = cfgGameName.Enabled = false;

            cfgGameProfile.Items.AddRange(Launcher.JavaProfiles.ToArray());
            if (Launcher.JavaProfiles.Count > 0)
                cfgGameProfile.SelectedIndex = 0;

            //BuildLocalHashes();
            //GetRemoteHashes();
            //RefreshFolders();
            cfgSettings.TabPages.Remove(cfgFilesPage);
            cfgSettings.TabPages.Remove(cfgLauncherPage);
        }
        #endregion

        #region METHODS
        private string SizeToString(int size)
        {
            if (size < 1024)
                return string.Format("{0}MB", size);
            else
                return string.Format("{0:0.000}GB", Math.Round(size / 1024.0, 3));
        }
        private void Wait(Action action)
        {
            Thread t = new Thread(() =>
            {
                this.InvokeSafe(() => psnSpinner.Visible = true);
                action();
                this.InvokeSafe(() => psnSpinner.Visible = false);
            });
            t.Name = "[UI] Background";
            t.IsBackground = true;
            t.Start();
        }
        public void BuildLocalHashes()
        {
            cfgFilesLocalBuild.Enabled = false;

            if (!Directory.Exists(gameDir))
                Directory.CreateDirectory(gameDir);

            this.pgbProgress.Visible = true;
            this.pgbProgress.Value = 0;

            Wait(() =>
            {
                string[] files = Directory.GetFiles(gameDir, "*.*", SearchOption.AllDirectories);
                this.InvokeSafe(() => pgbProgress.Maximum = files.Length);
                this.localHashes.Clear();
                for (int i = 0; i < files.Length; i++)
                {
                    string file = files[i];
                    string name = file.Replace(gameDir, "");
                    try { this.localHashes.Add(FileHashEntry.FromFile(name, file)); }
                    catch { this.localHashes.Add(new FileHashEntry(name, "-")); }
                    this.InvokeSafe(() => pgbProgress.Value = i);
                }
                this.InvokeSafe(() =>
                {
                    UpdateList();
                    cfgFilesLocalBuild.Enabled = true;
                    pgbProgress.Visible = false;
                });
            });
        }
        private void ShowError(string format, params object[] args)
        {
            this.InvokeSafe(() => MetroFramework.MetroMessageBox.Show(Launcher.MainForm, string.Format(format, args), "Fehler", MessageBoxButtons.OK, MessageBoxIcon.Error));
        }
        private void ShowSuccess(string format, params object[] args)
        {
            this.InvokeSafe(() => MetroFramework.MetroMessageBox.Show(Launcher.MainForm, string.Format(format, args), "MC Launcher v2", MessageBoxButtons.OK, MessageBoxIcon.Question));
        }
        public void GetRemoteHashes()
        {
            cfgFilesRemoteGet.Enabled = false;

            Wait(() =>
            {
                string content = null;
                try
                {
                    content = Encoding.UTF8.GetString(Launcher.FtpManager.DownloadFile("/hashes_game.json", null));
                } catch (Exception ex)
                {
                    ShowError("Konnte Server-Hashes nicht herunterladen:\n{0}", ex.Message);
                }

                if (!string.IsNullOrEmpty(content))
                {
                    List<FileHashEntry> remote = null;
                    try
                    {
                        remote = JsonConvert.DeserializeObject<List<FileHashEntry>>(content);
                    }catch(Exception ex)
                    {
                        ShowError("Konnte Server-Hashes nicht einlesen:\n{0}", ex.Message);
                    }
                    if (remote != null)
                        remoteHashes = remote;
                    else
                        remoteHashes.Clear();
                }
                this.InvokeSafe(() => {
                    UpdateList();
                    cfgFilesRemoteGet.Enabled = true;
                });
            });
        }
        public List<FileHashEntry> GetSelectedHashes()
        {
            List<FileHashEntry> hashes = new List<FileHashEntry>();
            foreach(ListViewItem item in ltvHashes.CheckedItems)
            {
                FileHashEntry[] e = item.Tag as FileHashEntry[];
                if (e != null)
                {
                    hashes.Add(e[0]);
                }
            }
            return hashes;
        }
        public void UpdateList()
        {
            List<ListViewItem> items = new List<ListViewItem>();
            foreach(var local in localHashes)
            {
                ListViewItem item = new ListViewItem();
                item.Text = local.Name;
                item.SubItems.Add(local.Hash);
                item.SubItems.Add("");
                item.SubItems.Add("");
                item.Tag = new FileHashEntry[] { local, null };
                items.Add(item);
            }
            foreach(var remote in remoteHashes)
            {
                if (items.Any(x=>x.Text == remote.Name))
                {
                    var item = items.First(x => x.Text == remote.Name);
                    item.SubItems[2].Text = remote.Hash;
                    FileHashEntry[] entries = item.Tag as FileHashEntry[];
                    item.Tag = new FileHashEntry[] { entries[0], remote };
                } else
                {
                    ListViewItem item = new ListViewItem();
                    item.Text = remote.Name;
                    item.SubItems.Add("");
                    item.SubItems.Add(remote.Hash);
                    item.SubItems.Add("");
                    item.Tag = new FileHashEntry[] { null, remote };
                    items.Add(item);
                }
            }

            //Check status
            foreach(var item in items)
            {
                FileHashEntry[] e = item.Tag as FileHashEntry[];
                if (e == null)
                    throw new Exception();

                string hashLocal = item.SubItems[1].Text;
                string hashRemote = item.SubItems[2].Text;
                
                if (hashLocal == "")
                    item.SubItems[3].Text = "Runterladen";
                else if (hashRemote == "")
                    item.SubItems[3].Text = "Hochladen";
                else if (hashLocal == hashRemote)
                    item.SubItems[3].Text = "<identisch>";
                else
                    item.SubItems[3].Text = "Konflikt";
            }

            //Check whitelist
            for (int i = items.Count - 1; i>= 0; i--)
            {
                if (items[i].Text.Count(x => x=='\\') > 1) //Not root
                {
                    if (!folderWhiteList.Any(x => x.Length > 0 && items[i].Text.StartsWith(x)))
                        items.RemoveAt(i);
                } else
                {
                    if (!folderWhiteList.Contains("\\"))
                        items.RemoveAt(i);
                }
            }

            var sorted = items.OrderBy(x => x.Text);

            ltvHashes.SuspendLayout();
            ltvHashes.Items.Clear();
            ltvHashes.Items.AddRange(sorted.ToArray());
            ltvHashes.ResumeLayout();
        }
        private void RefreshFolders()
        {
            if (!Directory.Exists(gameDir))
            {
                ltvFolders.Items.Clear();
                folderWhiteList.Clear();
            } else
            {
                var folders = Directory.GetDirectories(gameDir);
                List<ListViewItem> items = new List<ListViewItem>();
                foreach (var folder in folders.Concat(new string[] { "\\" }))
                {
                    ListViewItem item = new ListViewItem();
                    item.Text = folder.Replace(gameDir, "");
                    if (folderWhiteList.Contains(item.Text))
                        item.Checked = true;
                    items.Add(item);
                }
                //folderWhiteList.RemoveAll(x => !folders.Contains(x));

                ltvFolders.SuspendLayout();
                ltvFolders.Items.Clear();
                ltvFolders.Items.AddRange(items.OrderBy(x => x.Text).ToArray());
                ltvFolders.ResumeLayout();
            }
            RefreshBlacklist();
        }

        private void RefreshBlacklist()
        {
            ltvFolders.SuspendLayout();
            ltvFolders.Items.Clear();
            foreach (ListViewItem folder in ltvFolders.Items)
            {
                ListViewItem item = new ListViewItem();
                folder.Text = item.Text;
                folder.Checked = folderBlackList.Contains(item.Text);
                ltvFolders.Items.Add(item);
            }
            ltvFolders.ResumeLayout();
        }
        #endregion

        private void cfgGameProfile_SelectedIndexChanged(object sender, EventArgs e)
        {
            JavaProfile profile = cfgGameProfile.SelectedItem as JavaProfile;
            if (profile == null)
            {
                CurrentGameProfile = null;
            }
            else
            {
                CurrentGameProfile = profile;
            }
        }
        private void cfgGameSave_Click(object sender, EventArgs e)
        {
            if (currentGameProfile != null)
            {
                currentGameProfile.Args = cfgGameArgs.Text;
                currentGameProfile.Executable = cfgGameExecutable.Text;
                currentGameProfile.Name = cfgGameName.Text;
                currentGameProfile.Ram = cfgGameBar.Value;
                currentGameProfile.Version = (string)cfgGameVersion.SelectedItem ?? "";
                currentGameProfile.LibVersion = (string)cfgGameLibVersion.SelectedItem ?? "";
                Launcher.SaveJavaProfiles();
            }
        }
        private void metroButton2_Click(object sender, EventArgs e)
        {
            var profile = new JavaProfile();
            Launcher.JavaProfiles.Add(profile);
            cfgGameProfile.Items.Add(profile);
            cfgGameProfile.SelectedItem = profile;
        }
        private void cfgGameBar_ValueChanged(object sender, EventArgs e)
        {
            cfgGameRamCurrent.Text = SizeToString(cfgGameBar.Value);
        }
        private void cfgGameDelete_Click(object sender, EventArgs e)
        {
            if (currentGameProfile != null)
            {
                cfgGameProfile.Items.Remove(currentGameProfile);
                Launcher.JavaProfiles.Remove(currentGameProfile);
                CurrentGameProfile = null;
            }

        }

        private void cfgFilesLocalBuild_Click(object sender, EventArgs e)
        {
            BuildLocalHashes();
        }

        private void cfgFilesRemoteGet_Click(object sender, EventArgs e)
        {
            GetRemoteHashes();
        }

        private void alleAuswählenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in ltvHashes.Items)
                item.Checked = true;
        }

        private void auswahlUmkehrenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in ltvHashes.Items)
                item.Checked = !item.Checked;
        }

        private void auswahlAufhebenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in ltvHashes.Items)
                item.Checked = false;
        }

        private void metroButton3_Click(object sender, EventArgs e)
        {
            var hashes = GetSelectedHashes();
            cfgFilesUploadHashes.Enabled = false;

            Wait(() => {
                string json = JsonConvert.SerializeObject(hashes, Formatting.Indented);
                byte[] data = Encoding.UTF8.GetBytes(json);
                try
                {
                    using (MemoryStream str = new MemoryStream(data))
                    {
                        this.InvokeSafe(() => 
                        {
                            pgbProgress.Maximum = 100;
                            pgbProgress.Visible = true;
                        });
                        Launcher.FtpManager.UploadFile("/hashes_game.json", (p, t, c) =>
                        {
                            this.InvokeSafe(() => pgbProgress.Value = (int)Math.Max(100, p));
                        }, str, data.Length);
                        this.InvokeSafe(() => pgbProgress.Visible = false);
                    }
                    ShowSuccess("Hashes erfolgreich hochgeladen! ({0} bytes)", data.Length);
                }
                catch(Exception ex)
                {
                    ShowError("Konnte Hashes nicht hochladen:\n{0}", ex.Message);
                }
                this.InvokeSafe(() =>
                {
                    cfgFilesUploadHashes.Enabled = true;
                });
            });
        }

        private void cfgFilesFoldersRefresh_Click(object sender, EventArgs e)
        {
            RefreshFolders();
        }

        private void metroButton3_Click_1(object sender, EventArgs e)
        {
            SimpleJsonFile.SaveToFile("whitelist.json", folderWhiteList);
        }

        private void ltvFolders_ItemChecked(object sender, ItemCheckedEventArgs e)
        {
            if (!e.Item.Checked && folderWhiteList.Contains(e.Item.Text))
                folderWhiteList.Remove(e.Item.Text);
            else if (e.Item.Checked && !folderWhiteList.Contains(e.Item.Text))
                folderWhiteList.Add(e.Item.Text);

            UpdateList();
        }

        private void alleServerHashesAuswählenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            foreach(ListViewItem item in ltvHashes.Items)
            {
                FileHashEntry[] f = item.Tag as FileHashEntry[];
                if (f == null)
                    continue;

                if (remoteHashes.Contains(f[1]))
                    item.Checked = true;
            }
        }

        private void metroButton4_Click(object sender, EventArgs e)
        {
            SimpleJsonFile.SaveToFile("blacklist.json", folderBlackList);
        }

        private void cfgFilesDownload_Click(object sender, EventArgs e)
        {
            cfgFilesDownload.Enabled = false;
            pgbProgress.Value = 0;
            pgbProgress.Maximum = remoteHashes.Count;
            pgbProgress.Visible = true;

            pgbProgress2.Value = 0;
            pgbProgress2.Maximum = 100;
            pgbProgress2.Visible = true;

            Wait(() =>
            {
                List<FileHashEntry> failed = new List<FileHashEntry>();
                //Download new files
                for (int i = 0; i < remoteHashes.Count; i++)
                {
                    this.InvokeSafe(() => pgbProgress.Value = i);
                    bool download = false;
                    if (localHashes.Any(x => x.Name == remoteHashes[i].Name))
                    {
                        if (localHashes.First(x => x.Name == remoteHashes[i].Name).Hash != remoteHashes[i].Hash)
                            download = true;
                    }
                    else
                    {
                        download = true;
                    }
                    if (download)
                    {
                        this.InvokeSafe(() => pgbProgress2.Value = 0);
                        string path = gameDir + remoteHashes[i].Name;
                        FileInfo info = new FileInfo(path);
                        if (info.Exists)
                            info.Delete();
                        if (!info.Directory.Exists)
                            info.Directory.Create();
                        try
                        {
                            using (FileStream str = new FileStream(path, FileMode.CreateNew, FileAccess.ReadWrite))
                                Launcher.FtpManager.DownloadFile(remoteHashes[i].Name.Replace("\\", "/"), (p, t, c) =>
                                {
                                    this.InvokeSafe(() => pgbProgress2.Value = (int)Math.Max(100, p));
                                }, str);
                        }
                        catch { failed.Add(remoteHashes[i]); }
                    }
                }

                if (failed.Count > 0)
                {
                    string failedStr = string.Join("\n", failed.Select(x => x.Name).ToArray());
                    ShowError("Konnte folgende Datei(en) nicht herunterladen:\n{0}", failedStr);
                }
                this.InvokeSafe(() =>
                {
                    pgbProgress.Visible = false;
                    pgbProgress2.Visible = false;
                    cfgFilesDownload.Enabled = true;
                });
            });
        }

        private void cfgGameJavaExe_Click(object sender, EventArgs e)
        {
            using (OpenFileDialog diag = new OpenFileDialog())
            {
                diag.Title = "java.exe auswählen";
                diag.Filter = "Anwendung (*exe)|*.exe";
                if (File.Exists(cfgGameExecutable.Text))
                {
                    diag.FileName = new FileInfo(cfgGameExecutable.Text).Name;
                    diag.InitialDirectory = new FileInfo(cfgGameExecutable.Text).Directory.FullName;
                }
                if (diag.ShowDialog() == DialogResult.OK)
                {
                    cfgGameExecutable.Text = diag.FileName;
                }
            }
        }

        private void cfgGameJarFile_Click(object sender, EventArgs e)
        {
            //using (OpenFileDialog diag = new OpenFileDialog())
            //{
            //    diag.Title = "JAR-Datei auswählen";
            //    diag.Filter = "JAR (*.jar)|*.jar";
            //    if (File.Exists(cfgGameJarFile.Text))
            //    {
            //        diag.FileName = new FileInfo(cfgGameJarFile.Text).Name;
            //        diag.InitialDirectory = new FileInfo(cfgGameJarFile.Text).Directory.FullName;
            //    }
            //    if (diag.ShowDialog() == DialogResult.OK)
            //    {
            //        cfgGameJarFile.Text = diag.FileName;
            //    }
            //}
        }

        private void UpdateVersionEntries()
        {
            if (Directory.Exists(gameDir + "\\versions"))
            {
                var files = Directory.GetFiles(gameDir + "\\versions", "*.jar", SearchOption.AllDirectories).Select(x=>x.Replace(gameDir + "\\versions\\","")).ToArray();
                var folders = Directory.GetDirectories(gameDir + "\\versions", "*", SearchOption.TopDirectoryOnly).Select(x => x.Replace(gameDir + "\\versions\\", "")).ToArray();

                var selectedFile = cfgGameLibVersion.SelectedItem;
                var selectedFolder = cfgGameVersion.SelectedItem;

                cfgGameLibVersion.Items.Clear();
                cfgGameVersion.Items.Clear();

                cfgGameVersion.Items.AddRange(folders);
                if (folders.Contains(selectedFolder))
                    cfgGameVersion.SelectedItem = selectedFolder;
                else if (folders.Length > 0)
                    cfgGameVersion.SelectedItem = folders[0];

                cfgGameLibVersion.Items.AddRange(files);
                if (files.Contains(selectedFile))
                    cfgGameLibVersion.SelectedItem = selectedFile;
                else if (files.Length > 0)
                    cfgGameLibVersion.SelectedItem = files[0];
            }
            else
            {
                cfgGameVersion.Items.Clear();
                cfgGameLibVersion.Items.Clear();

                if (currentGameProfile != null)
                {
                    currentGameProfile.Version = "";
                    currentGameProfile.LibVersion = "";
                }
            }
        }

        private void cfgGameVersion_DropDown(object sender, EventArgs e)
        {
            UpdateVersionEntries();
        }

        private void UpdateProfileInfo()
        {
            btnGamePlay.Enabled = Launcher.JavaProfiles.Any(x => x.Active);
            if (btnGamePlay.Enabled)
                lblGameJavaProfile.Text = string.Format("Java-Profil: {0}", Launcher.JavaProfiles.First(x => x.Active));
            else
                lblGameJavaProfile.Text = string.Format("Java-Profil: Keines ausgewählt!");

            lblGameUserProfile.Text = "User-Profil: " + Launcher.SelectedUserProfile.DisplayName;
        }

        private void metroTabControl1_TabIndexChanged(object sender, EventArgs e)
        {
            UpdateProfileInfo();
        }

        private void btnGamePlay_Click(object sender, EventArgs e)
        {
            MinecraftManager m = new MinecraftManager();
            var info = m.BuildStartInfo(gameDir, Launcher.JavaProfiles.First(x => x.Active), Launcher.SelectedUserProfile);
            Process proc = new Process();
            proc.StartInfo = info;
            proc.Start();
            GameStarted?.Invoke(this, new EventArgs());
        }

        private void alleLokalenHashesAuswählenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in ltvHashes.Items)
            {
                FileHashEntry[] f = item.Tag as FileHashEntry[];
                if (f == null)
                    continue;

                if (localHashes.Contains(f[0]))
                    item.Checked = true;
            }
        }
    }
}
